/**
 * {模板解析, 基于 lit-html 二次开发}
 * @author yutent<yutent.io@gmail.com>
 * @date 2023/03/7 16:36:59
 */

import { boolMap, WC_PART, NOTHING } from './constants.js'
import { animate, MODES } from './anim.js'
import { nextTick, noop } from './utils.js'

const BIND_ATTR_SUFFIX = '{{$wkit$}}'
const MARKER = `{{^wkit${String(Math.random()).slice(-8)}^}}`
const MARKER_MATCH = '?' + MARKER
const NODE_MARKER = `<${MARKER_MATCH}>`

// 是否原始值
const isPrimitive = value =>
  value === null || (typeof value !== 'object' && typeof value !== 'function')
const isArray = Array.isArray
const isIterable = value =>
  value ? isArray(value) || typeof value[Symbol.iterator] === 'function' : false

const SPACE_CHAR = `[ \n\f\r]`
const ATTR_VALUE_CHAR = `[^ \n\f\r"'\`<>=]`
const NAME_CHAR = `[^\\s"'>=/]`
const TEXT_END_REGEX = /<(?:(!--|\/[^a-zA-Z])|(\/?[a-zA-Z][^>\s]*)|(\/?$))/g
const COMMENT_START = 1
const TAG_NAME = 2
const DYNAMIC_TAG_NAME = 3
const COMMENT_END_REGEXP = /-->/g
const COMMENT_END_REGEXP_2 = />/g
const TAG_END_REGEXP = new RegExp(
  `>|${SPACE_CHAR}(?:(${NAME_CHAR}+)(${SPACE_CHAR}*=${SPACE_CHAR}*(?:${ATTR_VALUE_CHAR}|("|')|))|$)`,
  'g'
)
const ENTIRE_MATCH = 0
const ATTRIBUTE_NAME = 1
const SPACES_AND_EQUALS = 2
const QUOTE_CHAR = 3
const SINGLE_QUOTE_REGEXP = /'/g
const DOUBLE_QUOTE_REGEXP = /"/g
const RAW_TEXT_ELEM_REGEXP = /^(?:script|style|textarea|title)$/i
const HTML_RESULT = 1
const SVG_RESULT = 2
const ATTRIBUTE_PART = 1
const CHILD_PART = 2
const EVENT_PART = 5
const ELEMENT_PART = 6
const COMMENT_PART = 7

const TEMPLATE_CACHE = new Map()
const walker = document.createTreeWalker(document, 129, null, false)

function getTemplateHtml(strings, type) {
  let len = strings.length - 1
  let attrNames = []
  let html2 = type === SVG_RESULT ? '<svg>' : ''
  let rawTextEndRegex
  let regex = TEXT_END_REGEX
  for (let i = 0; i < len; i++) {
    let s = strings[i]
    let attrNameEndIndex = -1
    let attrName
    let lastIndex = 0
    let match
    while (lastIndex < s.length) {
      regex.lastIndex = lastIndex
      match = regex.exec(s)
      if (match === null) {
        break
      }
      lastIndex = regex.lastIndex

      if (regex === TEXT_END_REGEX) {
        if (match[COMMENT_START] === '!--') {
          regex = COMMENT_END_REGEXP
        } else if (match[COMMENT_START] !== void 0) {
          regex = COMMENT_END_REGEXP_2
        } else if (match[TAG_NAME] !== void 0) {
          if (RAW_TEXT_ELEM_REGEXP.test(match[TAG_NAME])) {
            rawTextEndRegex = new RegExp(`</${match[TAG_NAME]}`, 'g')
          }
          regex = TAG_END_REGEXP
        } else if (match[DYNAMIC_TAG_NAME] !== void 0) {
          regex = TAG_END_REGEXP
        }
      } else if (regex === TAG_END_REGEXP) {
        if (match[ENTIRE_MATCH] === '>') {
          regex =
            rawTextEndRegex !== null && rawTextEndRegex !== void 0
              ? rawTextEndRegex
              : TEXT_END_REGEX
          attrNameEndIndex = -1
        } else if (match[ATTRIBUTE_NAME] === void 0) {
          attrNameEndIndex = -2
        } else {
          attrNameEndIndex = regex.lastIndex - match[SPACES_AND_EQUALS].length
          attrName = match[ATTRIBUTE_NAME]
          regex =
            match[QUOTE_CHAR] === void 0
              ? TAG_END_REGEXP
              : match[QUOTE_CHAR] === '"'
              ? DOUBLE_QUOTE_REGEXP
              : SINGLE_QUOTE_REGEXP
        }
      } else if (
        regex === DOUBLE_QUOTE_REGEXP ||
        regex === SINGLE_QUOTE_REGEXP
      ) {
        regex = TAG_END_REGEXP
      } else if (
        regex === COMMENT_END_REGEXP ||
        regex === COMMENT_END_REGEXP_2
      ) {
        regex = TEXT_END_REGEX
      } else {
        regex = TAG_END_REGEXP
        rawTextEndRegex = void 0
      }
    }

    let end =
      regex === TAG_END_REGEXP && strings[i + 1].startsWith('/>') ? ' ' : ''
    html2 +=
      regex === TEXT_END_REGEX
        ? s + NODE_MARKER
        : attrNameEndIndex >= 0
        ? (attrNames.push(attrName),
          s.slice(0, attrNameEndIndex) +
            BIND_ATTR_SUFFIX +
            s.slice(attrNameEndIndex)) +
          MARKER +
          end
        : s +
          MARKER +
          (attrNameEndIndex === -2 ? (attrNames.push(void 0), i) : end)
  }
  let htmlResult =
    html2 + (strings[len] || '<?>') + (type === SVG_RESULT ? '</svg>' : '')
  if (!isArray(strings) || !strings.hasOwnProperty('raw')) {
    throw new Error('invalid html ast')
  }
  return [htmlResult, attrNames]
}

function createElement(v = '') {
  let el = document.createElement('template')
  el.innerHTML = v
  return el
}

function createMarker(v = '') {
  return document.createComment(v)
}

class Template {
  parts = []

  constructor({ strings, values, __dom_type__: type }, options) {
    let node
    let nodeIndex = 0
    let attrNameIndex = 0
    let partCount = strings.length - 1
    let parts = this.parts
    let [html2, attrNames] = getTemplateHtml(strings, type)

    this.el = createElement(html2)

    walker.currentNode = this.el.content

    if (type === SVG_RESULT) {
      let content = this.el.content
      let svgElement = content.firstChild
      svgElement.remove()
      content.append(...svgElement.childNodes)
    }

    while ((node = walker.nextNode()) !== null && parts.length < partCount) {
      if (node.nodeType === 1) {
        if (node.hasAttributes()) {
          let attrsToRemove = []

          for (let name of node.getAttributeNames()) {
            if (name.endsWith(BIND_ATTR_SUFFIX) || name.startsWith(MARKER)) {
              let realName = attrNames[attrNameIndex++]
              attrsToRemove.push(name)
              if (realName !== void 0) {
                let value = node.getAttribute(
                  realName.toLowerCase() + BIND_ATTR_SUFFIX
                )
                let statics = value.split(MARKER)
                let m = /([#:@])?(.*)/.exec(realName)
                let decorates = []

                if (m[1] === '@' && m[2].includes('.')) {
                  decorates = m[2].split('.')
                  m[2] = decorates.shift()
                }

                parts.push({
                  type: ATTRIBUTE_PART,
                  index: nodeIndex,
                  name: m[2],
                  strings: statics,
                  decorates,
                  ctor:
                    m[1] === ':'
                      ? PropertyPart
                      : m[1] === '#' && m[2] === 'animation'
                      ? AnimPart
                      : m[1] === '@'
                      ? EventPart
                      : AttributePart
                })
              } else {
                parts.push({
                  type: ELEMENT_PART,
                  index: nodeIndex
                })
              }
            }
          }
          for (let name of attrsToRemove) {
            node.removeAttribute(name)
          }
        }
        if (RAW_TEXT_ELEM_REGEXP.test(node.tagName)) {
          let strings2 = node.textContent.split(MARKER)
          let lastIndex = strings2.length - 1
          if (lastIndex > 0) {
            node.textContent = ''
            for (let i = 0; i < lastIndex; i++) {
              node.append(strings2[i], createMarker())
              walker.nextNode()
              parts.push({ type: CHILD_PART, index: ++nodeIndex })
            }
            node.append(strings2[lastIndex], createMarker())
          }
        }
      } else if (node.nodeType === 8) {
        let data = node.data
        if (data === MARKER_MATCH) {
          parts.push({ type: CHILD_PART, index: nodeIndex })
        } else {
          let i = -1
          while ((i = node.data.indexOf(MARKER, i + 1)) !== -1) {
            parts.push({ type: COMMENT_PART, index: nodeIndex })
            i += MARKER.length - 1
          }
        }
      }
      nodeIndex++
    }
  }
}

class TemplateInstance {
  constructor(template, parent) {
    this.$parts = []
    this.$template = template
    this.$parent = parent
  }
  get parentNode() {
    return this.$parent.parentNode
  }

  #checkRef(node, walker, options) {
    do {
      if (node && node.nodeType === 1 && node.getAttribute('ref')) {
        options.host.$refs[node.getAttribute('ref')] = node
        node.removeAttribute('ref')
      }
    } while ((node = walker.nextNode()) !== null)
  }

  clone(options) {
    let {
      el: { content },
      parts
    } = this.$template

    let fragment = document.importNode(content, true)
    let nodeIndex = 0
    let partIndex = 0
    let templatePart = parts[0]
    let node = null

    walker.currentNode = fragment
    node = walker.nextNode()

    // 没有动态绑定时, 查一遍是否有ref属性
    if (templatePart === void 0) {
      this.#checkRef(node, walker, options)
    } else {
      while (templatePart !== void 0) {
        if (node.nodeType === 1 && node.getAttribute('ref')) {
          if (options.host.$refs) {
            options.host.$refs[node.getAttribute('ref')] = node
            node.removeAttribute('ref')
          }
        }
        if (nodeIndex === templatePart.index) {
          let part
          if (templatePart.type === CHILD_PART) {
            part = new ChildPart(node, node.nextSibling, this, options)
          } else if (templatePart.type === ATTRIBUTE_PART) {
            part = new templatePart.ctor(
              node,
              templatePart.name,
              templatePart.strings,
              templatePart.decorates,
              this,
              options
            )
          } else if (templatePart.type === ELEMENT_PART) {
            part = new ElementPart(node, this, options)
          }
          this.$parts.push(part)
          templatePart = parts[++partIndex]
        }
        if (
          nodeIndex !==
          (templatePart === null || templatePart === void 0
            ? void 0
            : templatePart.index)
        ) {
          node = walker.nextNode()
          nodeIndex++
        }
      }
      // 再检查剩下没有动态绑定的节点
      this.#checkRef(node, walker, options)
    }

    return fragment
  }

  update(values) {
    let i = 0
    for (let part of this.$parts) {
      if (part !== void 0) {
        if (part.strings !== void 0) {
          part.$setValue(values, i)
          i += part.strings.length - 2
        } else {
          part.$setValue(values[i])
        }
      }
      i++
    }
  }
}

class ChildPart {
  type = CHILD_PART
  #value = NOTHING

  constructor(startNode, endNode, parent, options) {
    this.startNode = startNode
    this.endNode = endNode
    this.$parent = parent
    this.options = options
  }

  get parentNode() {
    let parentNode = this.startNode.parentNode
    let parent = this.$parent
    if (parent !== void 0 && parentNode.nodeType === 11) {
      parentNode = parent.parentNode
    }
    return parentNode
  }

  $setValue(value, clear) {
    if (isPrimitive(value)) {
      if (value === NOTHING || value == null || value === '') {
        if (this.#value !== NOTHING) {
          this.#clear()
        }
        this.#value = NOTHING
      } else if (value !== this.#value) {
        this.#commitText(value)
      }
    } else if (value.__dom_type__ !== void 0) {
      this.#commitTemplateResult(value)
    } else if (value.nodeType !== void 0) {
      this.#commitNode(value)
    } else if (isIterable(value)) {
      this.#commitIterable(value)
    } else {
      this.#commitText(value)
    }
    if (clear) {
      nextTick(_ => this.#clearBindings())
    }
  }

  #insert(node, target = this.endNode) {
    return this.startNode.parentNode.insertBefore(node, target)
  }

  #commitNode(value) {
    if (this.#value !== value) {
      this.#clear()
      this.#value = this.#insert(value)
    }
  }

  #commitText(value) {
    if (this.#value !== NOTHING && isPrimitive(this.#value)) {
      let node = this.startNode.nextSibling

      node.data = value
    } else {
      this.#commitNode(document.createTextNode(value))
    }
    this.#value = value
  }

  #commitTemplateResult(result) {
    let { values, __dom_type__: type } = result
    let template =
      typeof type === 'number'
        ? this.#getTemplate(result)
        : (type.el === void 0 && (type.el = createElement(type.h)), type)

    if (this.#value?.$template === template) {
      this.#value.update(values)
    } else {
      let instance = new TemplateInstance(template, this)
      let fragment = instance.clone(this.options)

      instance.update(values)

      this.#commitNode(fragment)
      this.#value = instance
    }
  }

  #getTemplate(result) {
    let key = result.strings.join()
    let template = TEMPLATE_CACHE.get(key)

    if (template === void 0) {
      template = new Template(result, this.options)
      TEMPLATE_CACHE.set(key, template)
    }
    return template
  }

  #commitIterable(value) {
    if (!isArray(this.#value)) {
      this.#value = []
      this.#clear()
    }
    let itemParts = this.#value
    let partIndex = 0
    let itemPart
    for (let item of value) {
      if (partIndex === itemParts.length) {
        itemParts.push(
          (itemPart = new ChildPart(
            this.#insert(createMarker()),
            this.#insert(createMarker()),
            this,
            this.options
          ))
        )
      } else {
        itemPart = itemParts[partIndex]
      }
      itemPart.$setValue(item)
      partIndex++
    }
    if (partIndex < itemParts.length) {
      this.#clear(itemPart && itemPart.endNode.nextSibling, partIndex)
      itemParts.length = partIndex
    }
  }

  #clear(start = this.startNode.nextSibling, from) {
    while (start && start !== this.endNode) {
      let node = start.nextSibling
      start.remove()
      start = node
    }
  }

  /**
   * 清除已被移除的节点事件
   */
  #clearBindings() {
    let host = this.options.host
    let events = host.$events

    for (let ev in events) {
      for (let i = -1, it; (it = events[ev][++i]); ) {
        if (!host.root.contains(it.el)) {
          it.el.removeEventListener(this.name, it.listener, it.options)
          events[ev][i] = null
        }
      }
      events[ev] = events[ev].filter(e => e !== null)
    }
  }
}
// 常规属性
class AttributePart {
  type = ATTRIBUTE_PART

  #value = NOTHING

  constructor(element, name, strings, decorates, parent, options = {}) {
    this.element = element
    this.name = name
    this.decorates = decorates
    this.$parent = parent
    this.options = options
    if (strings.length > 2 || strings[0] !== '' || strings[1] !== '') {
      this.#value = Array(strings.length - 1).fill(null)
      this.strings = strings
    }
  }
  get tagName() {
    return this.element.tagName
  }

  $setValue(value, valueIndex) {
    let strings = this.strings
    let changed = false

    if (typeof value === 'function') {
      value = value(this.element)
    }

    if (strings === void 0) {
      changed = !isPrimitive(value) || value !== this.#value
      if (changed) {
        this.#value = value
      }
    } else {
      let values = value
      value = strings[0]

      for (let i = 0; i < strings.length - 1; i++) {
        let v = values[valueIndex + i]

        changed || (changed = !isPrimitive(v) || v !== this.#value[i])
        if (v === NOTHING) {
          value = NOTHING
        } else if (value !== NOTHING) {
          value += (v !== null && v !== void 0 ? v : '') + strings[i + 1]
        }
        this.#value[i] = v
      }
    }
    // value属性做特殊处理, 解决输入框无法清空的问题
    if (changed || this.name === 'value') {
      this.commitValue(value)
    }
  }

  commitValue(value) {
    let elem = this.element
    let attr = this.name
    let isBoolAttr = boolMap[attr]
    // ref属性不渲染到节点上
    if (attr === 'ref') {
      this.options.host.$refs[value] = elem
      return
    }

    if (isBoolAttr) {
      elem[isBoolAttr] = !(value === false || value === null)

      if (elem[isBoolAttr]) {
        elem.setAttribute(attr, '')
      } else {
        elem.removeAttribute(attr)
      }
    } else {
      // value 绑定, 特殊处理
      if (attr === 'value') {
        elem.value = value
      } else {
        if (value === null || value === void 0) {
          elem.removeAttribute(attr)
        } else {
          elem.setAttribute(attr, value)
        }
      }
    }
  }
}
// 赋值属性
class PropertyPart extends AttributePart {
  constructor(...args) {
    super(...args)
  }

  commitValue(value) {
    this.element[this.name] = value
  }
}
// 动画属性
class AnimPart extends AttributePart {
  constructor(...args) {
    super(...args)
  }

  commitValue({ type = 'fade', duration, custom, immediate = false } = {}) {
    let elem = this.element
    let fromto = MODES[type] || MODES.fade

    if (custom) {
      fromto = custom
    }
    elem.$animate = function (out = false) {
      return animate.call(this, duration, fromto, out)
    }
    elem.$animate.immediate = immediate
  }
}

// 事件属性
class EventPart extends AttributePart {
  type = EVENT_PART

  #listener = null

  #prevent = noop
  #stop = noop
  #checkSelf = noop

  constructor(...args) {
    super(...args)
  }

  $setValue(listener) {
    let host = this.options.host
    let elem = this.element
    let name = this.name
    let options = {}
    let events = host.$events[name] || []

    if (this.decorates.length) {
      for (let it of this.decorates) {
        switch (it) {
          case 'stop':
            this.#stop = ev => ev.stopPropagation()
            break
          case 'prevent':
            this.#prevent = ev => ev.preventDefault()
            break
          case 'self':
            this.#checkSelf = ev => ev.target === elem
            break
          case 'capture':
          case 'once':
          case 'passive':
            options[it] = true
            break
        }
      }
    }

    let shouldRemove = listener !== this.#listener

    if (this.#listener) {
      for (let i = -1, it; (it = events[++i]); ) {
        if (it.el === elem) {
          shouldRemove =
            shouldRemove ||
            options.capture !== it.capture ||
            options.once !== it.once ||
            options.passive !== it.passive
          if (shouldRemove) {
            it.el.removeEventListener(name, it.listener, it.options)
            events[i] = null
          }
          break
        }
      }
    }

    events = events.filter(it => it !== null)
    host.$events[name] = events

    if (listener && shouldRemove) {
      elem.addEventListener(name, this, options)
      this.#listener = listener

      events.push({ el: elem, listener: this, options })
    }
  }

  handleEvent(ev) {
    this.#stop(ev)
    this.#prevent(ev)
    if (this.#checkSelf(ev) === false) {
      return
    }
    this.#listener.call(this.options.host, ev)
  }
}

class ElementPart {
  type = ELEMENT_PART
  constructor(element, parent, options) {
    this.element = element
    this.$parent = parent
    this.options = options
  }

  $setValue() {}
}

export function render(value, container, options = {}) {
  let part = container[WC_PART]

  if (part === void 0) {
    part = new ChildPart(
      container.insertBefore(createMarker(), null),
      null,
      void 0,
      options
    )
    container[WC_PART] = part
  }
  part.$setValue(value, true)

  return part
}

export function html(strings, ...values) {
  return {
    __dom_type__: HTML_RESULT,
    strings,
    values
  }
}

export function raw(str, values = []) {
  let strings = values.length ? str.split('%s') : [str]
  strings.raw = true
  return html(strings, ...values)
}
export function svg(strings, ...values) {
  return {
    __dom_type__: SVG_RESULT,
    strings,
    values
  }
}
